/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II

	Notes:
	Quake script command precessing module
======================================================================*/

#include "quakedef.h"
#include "protocol.h"
#include "cmd.h"
#include "console.h"
#include "cvar.h"
#include "heap.h"
#include "host.h"
#include "sys.h"

using namespace Axiom;

extern cvar_t cmdline;

void Cmd_ForwardToServer();

cmdalias_t	*cmd_alias;
complist_t *complist = NULL;
static cmd_t *cmd_functions = NULL;

// numbers of cvars and cmds
int numcomplist = 0;

int trashtest;
int *trashspot;

bool	cmd_wait;

typedef struct cmd_function_s
{
	struct cmd_function_s	*next;
	char					*name;
	xcommand_t				function;
} cmd_function_t;

#define	MAX_ARGS		80

static	int			cmd_argc;
static	char		*cmd_argv[MAX_ARGS];
static	char		*cmd_null_string = "";
static	char		*cmd_args = NULL;

cmd_source_t	cmd_source;

namespace Axiom
{
	void Cmd_Compat() {} // Dummy function

	int CmdCvarCompareFunc(const void *a, const void *b)
	{
		complist_t *cc1 = (complist_t *) a;
		complist_t *cc2 = (complist_t *) b;

		return _stricmp (cc1->name, cc2->name);
	}

	cvar_t *Cmd_FindCvar (char *name)
	{
		complist_t key = {name, NULL, NULL, NULL};
		complist_t *cl = (complist_t *) bsearch (&key, complist, numcomplist, sizeof (complist_t), CmdCvarCompareFunc);

		if (!cl)
			return NULL;
		else if (cl->var)
		{
			return cl->var;
		}
		else return NULL;
	}

	void Cmd_BuildCompletionList()
	{
		// nothing to start with
		numcomplist = 0;

		// count the number of cvars and cmds we have
		for (cvar_t *var = cvar_vars; var; var = var->next) numcomplist++;
		for (cmd_t *cmd = cmd_functions; cmd; cmd = cmd->next) numcomplist++;
		for (cmdalias_t *ali = cmd_alias; ali; ali = ali->next) numcomplist++;
		for (cvar_alias_t *ali = cvar_alias_vars; ali; ali = ali->next) numcomplist++;

		// alloc space for the completion list (add some overshoot here; we need 1 to NULL terminate the list)
		// place in zone so that we can rebuild the list if we ever need to.
		if (complist) Zone_Free (complist);
		complist = (complist_t *) Zone_Alloc ((numcomplist + 1) * sizeof (complist_t));

		// current item we're working on
		complist_t *complistcurrent = complist;

		// write in cvars
		for (cvar_t *var = cvar_vars; var; var = var->next, complistcurrent++)
		{
			complistcurrent->name = var->name;
			complistcurrent->als = NULL;
			complistcurrent->cmd = NULL;
			complistcurrent->var = var;
		}

		// write in alias cvars
		for (cvar_alias_t *ali = cvar_alias_vars; ali; ali = ali->next, complistcurrent++)
		{
			complistcurrent->name = ali->name;
			complistcurrent->als = NULL;
			complistcurrent->cmd = NULL;
			complistcurrent->var = ali->var;
		}

		// write in cmds
		for (cmd_t *cmd = cmd_functions; cmd; cmd = cmd->next, complistcurrent++)
		{
			// replace NULL function commands with a dummy that does nothing
			if (!cmd->function)
			{
				cmd->usage = CMD_COMPAT;
				cmd->function = Cmd_Compat;
			}

			complistcurrent->name = cmd->name;
			complistcurrent->als = NULL;
			complistcurrent->cmd = cmd;
			complistcurrent->var = NULL;
		}

		// write in aliases
		for (cmdalias_t *als = cmd_alias; als; als = als->next, complistcurrent++)
		{
			complistcurrent->name = als->name;
			complistcurrent->als = als;
			complistcurrent->cmd = NULL;
			complistcurrent->var = NULL;
		}

		// sort before termination
		qsort ((void *) complist, numcomplist, sizeof (complist_t), CmdCvarCompareFunc);

		// terminate the list
		complistcurrent->name = NULL;
		complistcurrent->als = NULL;
		complistcurrent->cmd = NULL;
		complistcurrent->var = NULL;
	}

	void CmdCvarList (bool dumpcmd, bool dumpvar)
	{
		if (Cmd_Argc () == 1)
		{
			for (int i = 0; i < numcomplist; i++)
			{
				if (complist[i].cmd && dumpcmd) Con_Printf ("%s\n", complist[i].name);
				if (complist[i].var && dumpvar) Con_Printf ("%s\n", complist[i].name);
			}

			Con_Printf ("Use \"%s <filename>\" to dump to file\n", Cmd_Argv (0));
			return;
		}

		// because I just know somebody's gonna do this...
		char filename[MAX_PATH] = {0};

		for (int i = 1; i < Cmd_Argc (); i++)
		{
			// prevent a buffer overrun here
			if ((strlen (filename) + strlen (Cmd_Argv (i)) + 2) >= MAX_PATH) break;

			strcat (filename, Cmd_Argv (i));
			strcat (filename, " ");
		}

		for (int i = strlen (filename) - 1; i; i--)
		{
			if (filename[i] == ' ')
			{
				filename[i] = 0;
				break;
			}
		}

		FILE *f = fopen (filename, "w");

		if (!f)
		{
			Con_Printf ("Couldn't create \"%s\"\n", filename);
			return;
		}

		Con_Printf ("Dumping %s to \"%s\"... ", Cmd_Argv (0), filename);

		for (int i = 0; i < numcomplist; i++)
		{
			if (complist[i].cmd && dumpcmd) fprintf (f, "%s\n", complist[i].name);
			if (complist[i].var && dumpvar) fprintf (f, "%s\n", complist[i].name);
		}

		Con_Printf ("done\n");
		fclose (f);
	}

	void CmdList_f () {CmdCvarList (true, false);}
	void CvarList_f () {CmdCvarList (false, true);}
	void CmdCvarList_f () {CmdCvarList (true, true);}

	cmd_t CmdList_Cmd ("cmdlist", CmdList_f);
	cmd_t CvarList_Cmd ("cvarlist", CvarList_f);
	cmd_t CmdCvarList_Cmd ("cmdcvarlist", CmdCvarList_f);

	/*================================================================================
		Causes execution of the remainder of the command buffer to be delayed
		untill next frame. This allows commands like:
		bind g "impulse 5; +attack; wait; -attack; impulse 2"
	================================================================================*/
	void Cmd_Wait_f()
	{
		cmd_wait = true;
	}

	sizebuf_t	cmd_text;

	/*================================================================================
		Init
	================================================================================*/
	void Cbuf_Init()
	{
		SZ_Alloc (&cmd_text, 8192);		// space for commands and script files
	}

	/*================================================================================
		Adds command text at the end of the buffer
	================================================================================*/
	void Cbuf_AddText(char *text)
	{
		int		l;

		l = Q_strlen (text);

		if (cmd_text.cursize + l >= cmd_text.maxsize)
		{
			Con_Printf ("Cbuf_AddText: overflow\n");
			return;
		}

		SZ_Write (&cmd_text, text, Q_strlen (text));
	}

	/*================================================================================
		Adds command text immediatly after the current command
		Adds a \n to the text
		FIXME: Actually change the command buffer to do less copying
	================================================================================*/
	void Cbuf_InsertText(char *text)
	{
		char	*temp;
		int		templen;

		// copy off any commands still remaining in the exec buffer
		templen = cmd_text.cursize;
		if (templen)
		{
			temp = (char *)Z_Malloc (templen);
			Q_memcpy (temp, cmd_text.data, templen);
			SZ_Clear (&cmd_text);
		}
		else
			temp = NULL;	// shut up compiler

		// add the entire text of the file
		Cbuf_AddText (text);

		// add the copied off data
		if (templen)
		{
			SZ_Write (&cmd_text, temp, templen);
			Z_Free (temp);
		}
	}

	/*================================================================================
		Execute
	================================================================================*/
	void Cbuf_Execute()
	{
		int		i;
		char	*text;
		char	line[1024];
		int		quotes;

		while (cmd_text.cursize)
		{
			// find a \n or ; line break
			text = (char *)cmd_text.data;

			quotes = 0;
			for (i=0 ; i< cmd_text.cursize ; i++)
			{
				if (text[i] == '"')
					quotes++;
				if ( !(quotes&1) &&  text[i] == ';')
					break;	// don't break if inside a quoted string
				if (text[i] == '\n')
					break;
			}


			memcpy (line, text, i);
			line[i] = 0;

			// delete the text from the command buffer and move remaining commands down
			// this is necessary because commands (exec, alias) can insert data at the
			// beginning of the text buffer

			if (i == cmd_text.cursize)
				cmd_text.cursize = 0;
			else
			{
				i++;
				cmd_text.cursize -= i;
				Q_memcpy (text, text+i, cmd_text.cursize);
			}

			// execute the command line
			Cmd_ExecuteString (line, src_command);

			if (cmd_wait)
			{	// skip out while text still remains in buffer, leaving it
				// for next frame
				cmd_wait = false;
				break;
			}
		}
	}

	/*================================================================================
		Adds command line parameters as script statements
		Commands lead with a +, and continue until a - or another +
		quake +proj jctest.qp +cmd amlev1
		quake -nosound +cmd amlev1

		johnfitz -- Rewritten to read the "cmdline" cvar,
					for use with dynamic mod loading
	================================================================================*/
	void Cmd_StuffCmds_f()
	{
		char	cmds[CMDLINE_LENGTH];
		int		i, j, plus;

		plus = true;
		j = 0;
		for(i = 0; cmdline.string[i]; i++)
		{
			if (cmdline.string[i] == '+')
			{
				plus = true;
				if (j > 0)
				{
					cmds[j-1] = ';';
					cmds[j++] = ' ';
				}
			}
			else if (cmdline.string[i] == '-' &&
				(i==0 || cmdline.string[i-1] == ' ')) //johnfitz -- allow hypenated map names with +map
					plus = false;
			else if (plus)
				cmds[j++] = cmdline.string[i];
		}
		cmds[j] = 0;

		Cbuf_InsertText (cmds);
	}

	/*================================================================================
		Exec
	================================================================================*/
	void Cmd_Exec_f()
	{
		char	*f;
		int		mark;

		if (Cmd_Argc () != 2)
		{
			Con_Printf ("exec <filename> : execute a script file\n");
			return;
		}

		mark = Hunk_LowMark ();
		f = (char *)COM_LoadHunkFile (Cmd_Argv(1));
		if (!f)
		{
			Con_Printf ("couldn't exec %s\n",Cmd_Argv(1));
			return;
		}
		Con_Printf ("execing %s\n",Cmd_Argv(1));

		Cbuf_InsertText (f);
		Hunk_FreeToLowMark (mark);
	}

	/*================================================================================
		Just prints the rest of the line to the console
	================================================================================*/
	void Cmd_Echo_f()
	{
		int		i;

		for (i=1 ; i<Cmd_Argc() ; i++)
			Con_Printf ("%s ",Cmd_Argv(i));
		Con_Printf ("\n");
	}

	char *CopyString(char *in)
	{
		char	*out;

		out = (char *)Z_Malloc (strlen(in)+1);
		strcpy (out, in);
		return out;
	}

	/*================================================================================
		Creates a new command that executes a command string (possibly; seperated)
	================================================================================*/
	void Cmd_Alias_f()
	{
		cmdalias_t	*a;
		char		cmd[1024];
		int			i, c;
		char		*s;

		switch (Cmd_Argc())
		{
		case 1: //list all aliases
			for (a = cmd_alias, i = 0; a; a=a->next, i++)
				Con_SafePrintf ("   %s: %s", a->name, a->value);
			if (i)
				Con_SafePrintf ("%i alias command(s)\n", i);
			else
				Con_SafePrintf ("no alias commands found\n");
			break;
		case 2: //output current alias string
			for (a = cmd_alias ; a ; a=a->next)
				if (!strcmp(Cmd_Argv(1), a->name))
					Con_Printf ("   %s: %s", a->name, a->value);
			break;
		default: //set alias string
			s = Cmd_Argv(1);
			if (strlen(s) >= MAX_ALIAS_NAME)
			{
				Con_Printf ("Alias name is too long\n");
				return;
			}

			// if the alias allready exists, reuse it
			for (a = cmd_alias ; a ; a=a->next)
			{
				if (!strcmp(s, a->name))
				{
					Z_Free (a->value);
					break;
				}
			}

			if (!a)
			{
				a = (cmdalias_t *)Z_Malloc (sizeof(cmdalias_t));
				a->next = cmd_alias;
				cmd_alias = a;
			}
			strcpy (a->name, s);

			// copy the rest of the command line
			cmd[0] = 0;		// start out with a null string
			c = Cmd_Argc();
			for (i=2 ; i< c ; i++)
			{
				strcat (cmd, Cmd_Argv(i));
				if (i != c)
					strcat (cmd, " ");
			}
			strcat (cmd, "\n");

			a->value = CopyString (cmd);
			break;
		}
		Cmd_BuildCompletionList();
	}

	/*================================================================================
		Unalias
	================================================================================*/
	void Cmd_Unalias_f()
	{
		cmdalias_t	*a, *prev;

		switch (Cmd_Argc())
		{
		default:
		case 1:
			Con_Printf("unalias <name> : delete alias\n");
			break;
		case 2:
			for (prev = a = cmd_alias; a; a = a->next)
			{
				if (!strcmp(Cmd_Argv(1), a->name))
				{
					prev->next = a->next;
					Z_Free (a->value);
					Z_Free (a);
					prev = a;
					return;
				}
				prev = a;
			}
			break;
		}
	}

	/*================================================================================
		Unaliasall
	================================================================================*/
	void Cmd_Unaliasall_f()
	{
		cmdalias_t	*blah;

		while (cmd_alias)
		{
			blah = cmd_alias->next;
			Z_Free(cmd_alias->value);
			Z_Free(cmd_alias);
			cmd_alias = blah;
		}
	}

	/*================================================================================
		List
	================================================================================*/
	/*void Cmd_List_f()
	{
		cmd_function_t	*cmd;
		char 			*partial;
		int				len, count;

		if (Cmd_Argc() > 1)
		{
			partial = Cmd_Argv (1);
			len = Q_strlen(partial);
		}
		else
		{
			partial = NULL;
			len = 0;
		}

		count=0;
		for (cmd=cmd_functions ; cmd ; cmd=cmd->next)
		{
			if (partial && Q_strncmp (partial,cmd->name, len))
			{
				continue;
			}
			Con_SafePrintf ("   %s\n", cmd->name);
			count++;
		}

		Con_SafePrintf ("%i commands", count);
		if (partial)
		{
			Con_SafePrintf (" beginning with \"%s\"", partial);
		}
		Con_SafePrintf ("\n");
	}*/

	/*================================================================================
		Init
	================================================================================*/
	void Cmd_Init()
	{
		Cmd_BuildCompletionList();
	}
	
	// Feared; FIX
	//cmd_t Cmd_List_f_Cmd			("cmdlist",		Cmd_List_f);
	cmd_t Cmd_Unalias_f_Cmd			("unalias",		Cmd_Unalias_f);
	cmd_t Cmd_Unaliasall_f_Cmd		("unaliasall",	Cmd_Unaliasall_f);

	cmd_t Cmd_StuffCmds_f_Cmd		("stuffcmds",	Cmd_StuffCmds_f);
	cmd_t Cmd_Exec_f_Cmd			("exec",		Cmd_Exec_f);
	cmd_t Cmd_Echo_f_Cmd			("echo",		Cmd_Echo_f);
	cmd_t Cmd_Alias_f_Cmd			("alias",		Cmd_Alias_f);
	cmd_t Cmd_ForwardToServer_f_Cmd	("cmd",			Cmd_ForwardToServer);
	cmd_t Cmd_Wait_f_Cmd			("wait",		Cmd_Wait_f);

	/*================================================================================
		Argc
	================================================================================*/
	int Cmd_Argc()
	{
		return cmd_argc;
	}

	/*================================================================================
		Argv
	================================================================================*/
	char *Cmd_Argv(int arg)
	{
		if (arg >= cmd_argc)
			return cmd_null_string;
	
		return cmd_argv[arg];
	}

	/*================================================================================
		Args
	================================================================================*/
	char *Cmd_Args()
	{
		return cmd_args;
	}

	/*================================================================================
		Parses the given string into command line tokens.
	================================================================================*/
	void Cmd_TokenizeString(char *text)
	{
		int		i;

		// clear the args from the last string
		for (i=0 ; i<cmd_argc ; i++)
			Z_Free (cmd_argv[i]);

		cmd_argc = 0;
		cmd_args = NULL;

		while (1)
		{
			// skip whitespace up to a /n
			while (*text && *text <= ' ' && *text != '\n')
			{
				text++;
			}

			if (*text == '\n')
			{	// a newline seperates commands in the buffer
				text++;
				break;
			}

			if (!*text)
				return;

			if (cmd_argc == 1)
				 cmd_args = text;

			text = COM_Parse (text);
			if (!text)
				return;

			if (cmd_argc < MAX_ARGS)
			{
				cmd_argv[cmd_argc] = (char *)Z_Malloc (Q_strlen(com_token)+1);
				Q_strcpy (cmd_argv[cmd_argc], com_token);
				cmd_argc++;
			}
		}

	}

	/*================================================================================
		AddCommand
	================================================================================*/
	/*void cmd_t(char *cmd_name, xcommand_t function)
	{
		cmd_function_t	*cmd;
		cmd_function_t	*cursor,*prev; //johnfitz -- sorted list insert

		if (host_initialized)	// because hunk allocation would get stomped
			Sys_Error ("cmd_t after host_initialized");

		// fail if the command is a variable name
		if (Cvar_VariableString(cmd_name)[0])
		{
			Con_Printf ("cmd_t: %s already defined as a var\n", cmd_name);
			return;
		}

		// fail if the command already exists
		for (cmd = cmd_functions ; cmd ; cmd=cmd->next)
		{
			if (!Q_strcmp (cmd_name, cmd->name))
			{
				Con_Printf ("cmd_t: %s already defined\n", cmd_name);
				return;
			}
		}

		cmd = (cmd_function_t *)Hunk_Alloc (sizeof(cmd_function_t));
		cmd->name = cmd_name;
		cmd->function = function;

		//johnfitz -- insert each entry in alphabetical order
		if (cmd_functions == NULL || strcmp(cmd->name, cmd_functions->name) < 0) //insert at front
		{
			cmd->next = cmd_functions;
			cmd_functions = cmd;
		}
		else //insert later
		{
			prev = cmd_functions;
			cursor = cmd_functions->next;
			while ((cursor != NULL) && (strcmp(cmd->name, cursor->name) > 0))
			{
				prev = cursor;
				cursor = cursor->next;
			}
			cmd->next = prev->next;
			prev->next = cmd;
		}
		//johnfitz
	}*/

	/*================================================================================
		Exists
	================================================================================*/
	bool Cmd_Exists(char *cmd_name)
	{
		cmd_t	*cmd;

		for (cmd=cmd_functions ; cmd ; cmd=cmd->next)
		{
			if (!Q_strcmp (cmd_name,cmd->name))
				return true;
		}

		return false;
	}

	/*================================================================================
		CompleteCommand
	================================================================================*/
	char *Cmd_CompleteCommand(char *partial)
	{
		cmd_t	*cmd;
		int				len;

		len = Q_strlen(partial);

		if (!len)
			return NULL;

		// check functions
		for (cmd=cmd_functions ; cmd ; cmd=cmd->next)
			if (!Q_strncmp (partial,cmd->name, len))
				return cmd->name;

		return NULL;
	}

	/*================================================================================
		A complete command line has been parsed, so try to execute it
		FIXME: lookupnoadd the token to speed search?
	================================================================================*/
	void Cmd_ExecuteString(char *text, cmd_source_t src)
	{
		cmd_t		*cmd;
		cmdalias_t	*a;

		cmd_source = src;
		Cmd_TokenizeString (text);

		// execute the command line
		if (!Cmd_Argc())
			return;		// no tokens

		// check functions
		for (cmd=cmd_functions ; cmd ; cmd=cmd->next)
		{
			if (!Q_strcasecmp (cmd_argv[0],cmd->name))
			{
				cmd->function ();
				return;
			}
		}

		// check alias
		for (a=cmd_alias ; a ; a=a->next)
		{
			if (!Q_strcasecmp (cmd_argv[0], a->name))
			{
				Cbuf_InsertText (a->value);
				return;
			}
		}

		// check cvars
		if (!Cvar_Command ())
			Con_Printf ("Unknown command/cvar \"%s\"\n", Cmd_Argv(0));

	}

	/*================================================================================
		Sends the entire command line over to the server
	================================================================================*/
	void Cmd_ForwardToServer()
	{
		if (cls.state != ca_connected)
		{
			Con_Printf ("Can't \"%s\", not connected\n", Cmd_Argv(0));
			return;
		}

		if (cls.demoplayback)
			return;		// not really connected

		MSG_WriteByte (&cls.message, clc_stringcmd);
		if (Q_strcasecmp(Cmd_Argv(0), "cmd") != 0)
		{
			SZ_Print (&cls.message, Cmd_Argv(0));
			SZ_Print (&cls.message, " ");
		}
		if (Cmd_Argc() > 1)
			SZ_Print (&cls.message, Cmd_Args());
		else
			SZ_Print (&cls.message, "\n");
	}

	/*================================================================================
		Returns the position (1 to argc-1) in the command's argument list
		where the given parameter appears, or 0 if not present
	================================================================================*/
	int Cmd_CheckParm(char *parm)
	{
		int i;

		if (!parm)
			Sys_Error ("Cmd_CheckParm: NULL");

		for (i = 1; i < Cmd_Argc (); i++)
			if (! Q_strcasecmp (parm, Cmd_Argv (i)))
				return i;

		return 0;
	}

	void Cmd_Add (cmd_t *newcmd)
	{
		// fail if the command is a variable name
		if (Cvar_VariableString (newcmd->name)[0]) return;

		// fail if the command already exists
		for (cmd_t *cmd = cmd_functions; cmd; cmd = cmd->next)
		{
			if (!strcmp (newcmd->name, cmd->name))
			{
				// silent fail
				return;
			}
		}

		if (!newcmd->function)
		{
			newcmd->function = Cmd_Compat;
			newcmd->usage = CMD_COMPAT;
		}
		else newcmd->usage = CMD_NORMAL;

		// link in
		newcmd->next = cmd_functions;
		cmd_functions = newcmd;
	}

	cmd_t::cmd_t(char *cmdname, xcommand_t cmdcmd)
	{
		this->name = (char *) Zone_Alloc (strlen (cmdname) + 1);

		strcpy (this->name, cmdname);

		if (cmdcmd)
		{
			this->function = cmdcmd;
			this->usage = CMD_NORMAL;
		}
		else
		{
			this->function = Cmd_Compat;
			this->usage = CMD_COMPAT;
		}

		// just add it
		Cmd_Add (this);
	}
}